--
-- Winning
--https://i.imgur.com/EGCnB39.gif
-- Track battleground wins.
--

local C = Winning_Constants

local Graph = LibStub:GetLibrary("LibGraph-2.1")

local addon_name = "Winning"
local chat = Winning_Chat
local log = Winning_Log

local bg_record = WinningBattlegroundRecord

Winning = LibStub("AceAddon-3.0"):NewAddon(addon_name, "AceEvent-3.0", "AceConsole-3.0")
local addon = Winning
addon.version = C.ADDON_VERSION
addon.time_context_items = {
    C.PVP_TIME_ALL,
    C.PVP_TIME_SEASON,
    C.PVP_TIME_SIX_HOURS,
}
addon.type_context_items = {
    C.PVP_TYPE_ALL,
    C.PVP_TYPE_CASUAL,
    C.PVP_TYPE_RATED,
}

addon.map_items = {} -- idx -> BG Name
addon.spec_items = {}
addon.all_specs = {}
addon.ui_tabs = {'RECORDS', 'STATS'}

addon.ScrollingTable = LibStub("ScrollingTable");

local L = LibStub("AceLocale-3.0"):GetLocale(addon_name)
local addonLDB = LibStub("LibDataBroker-1.1"):NewDataObject(addon_name, {
	type = "data source",
	text = addon_name,
    icon = "Interface\\PvPRankBadges\\PvPRank12",
	OnClick = function(self, btn) Winning_Minimap_Click(btn) end,
});
local icon = LibStub("LibDBIcon-1.0")
local grey_cell_bg = { ["r"] = 0.15, ["g"] = 0.15, ["b"] = 0.15, ["a"] = 1.0 }



-------------------------------------------------------------------------------
-- Data Structures For Scrolling Tables
-------------------------------------------------------------------------------

addon.records_cols = {
    {
        ["name"] = L["Date"],
        ["width"] = 150,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
        ["comparesort"] = function (self, rowa, rowb, sortby)
            local rowa = self:GetRow(rowa);
            local rowb = self:GetRow(rowb);

            -- God damn this lib sucks.
            local column = self.cols[sortby]
            local sort_dir = column.sort or column.defaultsort or "asc"

            if sort_dir == 'asc' then
                return rowa[1] > rowb[1]
            else
                return rowa[1] < rowb[1]
            end

        end,
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            cellFrame.text:SetText(date("%a %H:%M %x", record[1]));
        end
    },{
        ["name"] = L["Map"],
        ["width"] = 150,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Duration"],
        ["width"] = 60,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            local value = tonumber(record[3])
            if value then
                cellFrame.text:SetText(addon:render_bg_duration(tonumber(value)))
            else
                cellFrame.text:SetText("")
            end
        end
    },{
        ["name"] = L["Winner"],
        ["width"] = 60,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["KBs"],
        ["width"] = 40,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["HKs"],
        ["width"] = 40,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Deaths"],
        ["width"] = 50,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Damage"],
        ["width"] = 60,
        ["align"] = "RIGHT",
        ["defaultsort"] = "dsc",
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            local value = record[8]
            cellFrame.text:SetText(addon:render_damage(value))
        end
    },
    {
        ["name"] = L["Spec"],
        ["width"] = 30,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            local value = record[9]
            local id, name, description, icon, background, role, class = GetSpecializationInfoByID(value)

            -- cellFrame.text:SetText(name)
            if cellFrame.Icon == nil then
                cellFrame.Icon = cellFrame:CreateTexture()
            end
            cellFrame.Icon:SetTexture(icon)
            cellFrame.Icon:SetWidth(15)
            cellFrame.Icon:SetHeight(15)
            cellFrame.Icon:SetPoint("TOP")
            cellFrame.Icon:Show()
        end,
        ["comparesort"] = function (self, rowa, rowb, sortby)
            local rowa = self:GetRow(rowa);
            local rowb = self:GetRow(rowb);

            -- God damn this lib sucks.
            local column = self.cols[sortby]
            local sort_dir = column.sort or column.defaultsort or "asc"

            if sort_dir == 'asc' then
                return rowa[9] > rowb[9]
            else
                return rowa[9] < rowb[9]
            end

        end,
    },
    {
        ["name"] = L["Healing"],
        ["width"] = 60,
        ["align"] = "RIGHT",
        ["defaultsort"] = "dsc",
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            local value = record[10]
            cellFrame.text:SetText(addon:render_damage(value))
        end
    },
    {
        ["name"] = L["Premades"],
        ["width"] = 60,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            local value = record[11]
            if value then
                cellFrame.text:SetText(string.format("%s", value['total']))
            else
                cellFrame.text:SetText('-')
            end

            -- I thought the enemy icon would be cool, but... it doesn't add much.
            --[[ cellFrame.text:SetText(name)
            if cellFrame.Icon == nil then
                cellFrame.Icon = cellFrame:CreateTexture()
            end
            if horde then
                cellFrame.Icon:SetTexture("Interface\\FriendsFrame\\PlusManz-Alliance")
            else
                cellFrame.Icon:SetTexture("Interface\\FriendsFrame\\PlusManz-Alliance")
            end
            cellFrame.Icon:SetWidth(15)
            cellFrame.Icon:SetHeight(15)
            cellFrame.Icon:SetPoint("TOP")
            cellFrame.Icon:Show()
            --]]
        end
    }
}

for i, col in pairs(addon.records_cols) do
    if i % 2 == 0 then
        col['bgcolor'] = grey_cell_bg
    end
end


-------------------------------------------------------------------------------
-- ACE Framework special methods.
-------------------------------------------------------------------------------

function addon:OnInitialize()
    log("Addon initializing.")

    self.db = LibStub("AceDB-3.0"):New(addon_name .. "Db", {
		profile = {
			minimap = {
				hide = false,
			},
		},
        char = {
            -- Table used for holding data for the UI
            bg_records = {},

            -- Master record of all BG data.
            all_records = {},

            -- Table used for calculating currently visible stats.
            bg_db_data = {},
        },
        faction = {
            -- Table used for holding data for the UI
            bg_records = {},

            -- Master record of all BG data.
            all_records = {},

            -- Table used for calculating currently visible stats.
            bg_db_data = {},

            type_context_value = "All",
            time_context_value = "All",
            map_context_value = "All",
            spec_context_value = "All",
            graph_context_value = 1,
        },
        global = {
            data_version = 0,
        }
	});

    self.dal = WinningDataAccessLayer:init(self)
    self.wiper = WinningWiper:init(self)

    self:setup_player_faction()
    self:setup_ui()

	icon:Register(addon_name, addonLDB, self.db.profile.minimap)
	self:RegisterChatCommand("winning", "slash_command")
end


function addon:OnEnable()
    self:RegisterEvent("ADDON_LOADED")
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:listen_for_bg_events()
    self:setupDisplay()
end


function addon:OnDisable()
    Winning_UI:Hide()
    self:ignore_bg_events()
end


-- Events

function addon:ADDON_LOADED()
    self:migrate()
    self:setupSpecs()
    self:setupDisplay()
end


function addon:PLAYER_ENTERING_WORLD()
    -- TODO: Don't do this if the addon is 'disabled' in config.
    self:listen_for_bg_events()
end


function addon:listen_for_bg_events()
    self:RegisterEvent("UPDATE_BATTLEFIELD_SCORE")
    self:RegisterEvent("UPDATE_BATTLEFIELD_STATUS")
    self:release_process_token()
end


function addon:ignore_bg_events()
    self:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE")
    self:UnregisterEvent("UPDATE_BATTLEFIELD_STATUS")
end


function addon:UPDATE_BATTLEFIELD_STATUS()
    log("UPDATE_BATTLEFIELD_STATUS")
    self:record_battleground_stats_if_over()
end


function addon:UPDATE_BATTLEFIELD_SCORE()
    log("UPDATE_BATTLEFIELD_SCORE")
    self:record_battleground_stats_if_over()
end



-------------------------------------------------------------------------------
-- Addon functionality
-------------------------------------------------------------------------------

function addon:slash_command(args)
    log("Slash command called: " .. tostring(args))

    if args == 'minimap' then
        self.db.profile.minimap.hide = not self.db.profile.minimap.hide
        if self.db.profile.minimap.hide then
            icon:Hide(addon_name)
        else
            icon:Show(addon_name)
        end
    elseif args == 'debug' then
        C.ADDON_DEBUG = not C.ADDON_DEBUG
    elseif args == 'stubs' then
        for i = 1, 50 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs5' then
        for i = 1, 5 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs500' then
        for i = 1, 500 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs5000' then
        for i = 1, 5000 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'wipe-mock' then
        self.wiper:wipe(addon.wiper.MOCK, true)
    elseif args == 'wipe-all' then
        self.wiper:wipe(addon.wiper.ALL, true)
    elseif args == 'wipe-season' then
        self.wiper:wipe(addon.wiper.SEASON, current_season)
    elseif args == 'wipe-previous' then
        self.wiper:wipe(addon.wiper.PREVIOUS_SEASONS, current_season)
    elseif args == 'resync' then
        self.dal:resync(self.wiper)
    elseif args == 'toggle' then
        Winning_Toggle_UI()
    elseif args == 'reset_defaults' then
        self.db.faction.type_context_value = "All"
        self.db.faction.time_context_value = "All"
        self.db.faction.map_context_value = "All"
        self.db.faction.spec_context_value = "All"
    elseif args == 'db_stats' then
        log("Char bg records: " .. tostring(#self.db.char.bg_records))
        log("Char all records: " .. tostring(#self.db.char.all_records))
        log("Char bg db records: " .. tostring(#self.db.char.bg_db_data))

        log("Faction bg records: " .. tostring(#self.db.faction.bg_records))
        log("Faction all records: " .. tostring(#self.db.faction.all_records))
        log("Faction bg db records: " .. tostring(#self.db.faction.bg_db_data))
        log("Global data version: " .. tostring(self.db.global.data_version))
    else

        chat("Available commands: minimap, debug, wipe-all, toggle" )
    end

    self:refresh()
end



function addon:request_process_token()
    if self._bg_process_token == nil then
        self:ignore_bg_events()
        self._bg_process_token = time()
    end
    return self._bg_process_token
end



function addon:release_process_token()
    self._bg_process_token = nil
end



--
-- Record Stats If Over
--
function addon:record_battleground_stats_if_over()
	local _, zone_type = IsInInstance()
    if zone_type == "pvp" then
        local bg_winning_faction = GetBattlefieldWinner()
        if bg_winning_faction ~= nil then
            if self:request_process_token() ~= nil then
                -- If saving a BG record breaks, the addon should still function
                local player_name = UnitName("player")
                local success, result = pcall(self.process_battleground_stats, self, player_name)
                if success then
                    log("Successfully recorded battleground stats.")
                else
                    log("Error recording battleground stats.")
                    log(tostring(result))
                end
            else
                log("Match over, but could not get token. Ignoring.")
            end
        else
            log("Match not over yet.")
        end
    end
end


--
-- Collect BG Stats for current player, for current BG.
--
-- Depends on player still being in the BG.
-- TODO: What happens if the player rage quits/leaves early?
--
function addon:process_battleground_stats(player_name)
    bg = WinningBgRecord:new(player_name)
    local bg_record = {}
    bg_record.uuid = Winning_UUID()
    bg_record.PlayerName = bg:get_player_name()
    bg_record.PlayerSpec = bg:get_player_spec()
    bg_record.PlayerRace = bg:get_player_race()
    bg_record.Type = 'bg_record'
    bg_record.Date = bg:get_date()
    bg_record.MapId = bg:get_map_id()
    bg_record.Map = self:render_bg_map(bg_record.MapId)
    bg_record.DurationRaw = bg:get_duration()
    bg_record.Duration = self:render_bg_duration(bg_record.DurationRaw)
    bg_record.Won = bg:get_won()
    bg_record.Winner = self:render_bg_won(bg_record.Won)
    bg_record.KBs = bg:get_kbs()
    bg_record.HKs = bg:get_hks()
    bg_record.Deaths = bg:get_deaths()
    bg_record.DamageRaw = bg:get_player_damage()
    bg_record.Damage = self:render_damage(bg_record.DamageRaw)
    bg_record.HealingRaw = bg:get_player_healing()
    bg_record.Healing = self:render_healing(bg_record.HealingRaw)
    bg_record.Season = bg:get_season()
    bg_record.Rated = bg:get_was_rated_match()
    bg_record.AllyPremades = bg:get_ally_premades(self.player_faction_num)
    bg_record.EnemyPremades = bg:get_enemy_premades(self.player_faction_num)
    self.dal:save(bg_record)
end


function addon:mock_bg_record(i)
    local map = math.random(2, 13)
    local map_name = GetBattlegroundInfo(map)
    local winner = math.random(0, 1)
    local won = winner == self.player_faction_num
    local duration = math.random(300, 1200)
    local damage = math.random(100000, 9999999)
    local healing = math.random(100000, 9999999)

    local _, race = UnitRace("player")

    return {
        uuid = Winning_UUID(),
        PlayerName = 'foo' .. tostring(i),
        PlayerSpec = 102, -- balance druid
        PlayerRace = race,
        Type = 'bg_record',
        Date = time() + math.random(1000, (1000*i)+1) - 1000000000,
        MapId = map,
        Map = map_name,
        DurationRaw = duration,
        Duration = self:render_bg_duration(duration),
        Won = won,
        Winner = addon:render_bg_won(won),
        KBs = math.random(1, 30),
        HKs = math.random(1, 100),
        Deaths = math.random(1, 10),
        Damage = self:render_damage(damage),
        DamageRaw = damage,
        Healing = self:render_healing(healing),
        HealingRaw = healing,
        Season = 1,
        Rated = false,
    }
end


-- Used for rendering bg record data.
function addon:setup_player_faction()
    self.player_faction = UnitFactionGroup("player");
    if self.player_faction == "Neutral" then
        self.player_faction_num = nil
    else
        self.player_faction_num = 1
        if self.player_faction == "Horde" then
            self.player_faction_num = 0
        end
    end
    log("Player faction is: " .. self.player_faction)
    log("Player faction num is: " .. tostring(self.player_faction_num))
end


function addon:render_bg_won(won)
    -- Horde team always Red
    -- Alliance team always Blue
    -- text for opposing team depends on what faction the player was playing as

    local alliance = self:blue(FACTION_ALLIANCE)
    local horde = self:red(FACTION_HORDE)

    if won then
        if self.player_faction_num == 0 then
            return horde
        else
            return alliance
        end
    else
        -- if the player did not win, always return opposite faction in that
        -- faction's color.
        if self.player_faction_num == 0 then
            return alliance
        else
            return horde
        end
    end
end


function addon:render_bg_duration(total_seconds)
    return Winning_Duration(total_seconds)
end


function addon:blue(txt)
    return format("%s%s\124r", C.COLORS['blue'], txt)
end

function addon:green(txt)
    return format("%s%s\124r", C.COLORS['green'], txt)
end

function addon:red(txt)
    return format("%s%s\124r", C.COLORS['red'], txt)
end


function addon:render_bg_map(map)
    return GetMapNameByID(map)
end


function addon:render_damage(damage)
    return Winning_Humanify(damage)
end

function addon:render_healing(healing)
    return Winning_Humanify(healing)
end



-------------------------------------------------------------------------------
-- UI Functions
-------------------------------------------------------------------------------

-- one-time setup of the UI
function addon:setup_ui()
    log("Setting up UI for one time elements.")

    Winning_UI_Title:SetText(format("%s v%s", addon_name, addon.version))

    -- Content Tab Labels
    Winning_UITab1:SetText(L['Records'])
    Winning_UITab2:SetText(L['Stats'])

    -- Records Tab
    Winning_UI_Info_Context_Type_Label:SetText(L['Type'])
    Winning_UI_Info_Context_Time_Label:SetText(L['Context'])
    Winning_UI_Info_Context_Maps_Label:SetText(L['Maps'])
    Winning_UI_Info_Context_Spec_Label:SetText(L['Spec'])

    -- setup context dropdown
    UIDropDownMenu_Initialize(Winning_Type_DropDown, Winning_Type_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Type_DropDown, 100);
    UIDropDownMenu_SetButtonWidth(Winning_Type_DropDown, 100);
    UIDropDownMenu_SetSelectedID(Winning_Type_DropDown, addon.db.char.type_context_value or 1)
    UIDropDownMenu_JustifyText(Winning_Type_DropDown, "LEFT");

    UIDropDownMenu_Initialize(Winning_Context_DropDown, Winning_Context_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Context_DropDown, 100);
    UIDropDownMenu_SetButtonWidth(Winning_Context_DropDown, 100);
    UIDropDownMenu_SetSelectedID(Winning_Context_DropDown, addon.db.char.time_context_value or 1)
    UIDropDownMenu_JustifyText(Winning_Context_DropDown, "LEFT");

    UIDropDownMenu_Initialize(Winning_Map_DropDown, Winning_Map_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Map_DropDown, 140);
    UIDropDownMenu_SetButtonWidth(Winning_Map_DropDown, 100);
    -- Use name here because the infos for the dropdown don't necessarily exist yet, or something.
    --UIDropDownMenu_SetSelectedName(Winning_Map_DropDown, addon.map_items[addon.db.char.map_context_value] or "All")
    UIDropDownMenu_JustifyText(Winning_Map_DropDown, "LEFT");

    UIDropDownMenu_Initialize(Winning_Spec_DropDown, Winning_Spec_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Spec_DropDown, 140);
    UIDropDownMenu_SetButtonWidth(Winning_Spec_DropDown, 100);
    UIDropDownMenu_JustifyText(Winning_Spec_DropDown, "LEFT");

    -- context dropdown for graph details
    UIDropDownMenu_Initialize(Winning_Graph_Context_DropDown, Winning_Graph_Context_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Graph_Context_DropDown, 100);
    UIDropDownMenu_SetButtonWidth(Winning_Graph_Context_DropDown, 100);
    UIDropDownMenu_SetSelectedID(Winning_Graph_Context_DropDown, 1);
    UIDropDownMenu_JustifyText(Winning_Graph_Context_DropDown, "LEFT");


    Winning_UI_Info_Ratio_Label:SetText(L['WINNING'])

    Winning_UI_Info_Stats_HKs_Label:SetText(L['Top HKs'])
    Winning_UI_Info_Stats_KBs_Label:SetText(L['Top KBs'])
    Winning_UI_Info_Stats_Damage_Label:SetText(L['Damage'])
    Winning_UI_Info_Stats_Healing_Label:SetText(L['Healing'])
    Winning_UI_Info_Stats_Streak_Label:SetText(L['Streak'])

    Winning_Info_Report_Label:SetText(L['Report'])
    Winning_UI_Report_Party:SetText(L['Party'])
    Winning_UI_Report_Guild:SetText(L['Guild'])
    Winning_UI_Report_Premades:SetText(L['Premades'])

    if self.records_table == nil then
        self.records_table = addon.ScrollingTable:CreateST(addon.records_cols, 25, nil, nil, Winning_UI_Records_Table)
        self.records_table:SetData(self.dal.ui_data, true)
    end
    _G[self.records_table["frame"]:GetName()]:SetAllPoints()


    -- Set up the stats graph.
    -- TODO: localize labels
    -- TODO: Start based on preferred graph context

    Winning_UI_Stats_Maps_Details_Graph_1_Label:SetText(L['Daily Winning Percentage'])

	local g = Graph:CreateGraphLine("Winning_Stats_Graph_1", Winning_UI_Stats_Maps_Details_Graph_1)
    addon.stats_graph = g

    local xLabels = {}
    xLabels[1] = "Sun"
    xLabels[2] = "Mon"
    xLabels[3] = "Tues"
    xLabels[4] = "Wed"
    xLabels[5] = "Thurs"
    xLabels[6] = "Fri"
    xLabels[7] = "Sat"

	g:SetXAxis(0, 8)
    g:SetXAxisLabels(xLabels)

	g:SetYAxis(0, 100)
    g:SetYLabels(false, false)
	g:SetGridSpacing(1, 50)
	g:SetGridColor({0.5, 0.5, 0.5, 0.1})
	g:SetAxisDrawing(true, true)
	g:SetAxisColor({0.75, 0.75, 0.75, 0.75})
	g:SetAutoScale(true)
	g:LockYMin(true)

	local Data1 = {
        {1, 1},
        {2, 3},
        {3, 7},
        {4, 0},
        {5, 3},
        {6, 9},
        {7, 14},
    }

	g:AddFilledDataSeries(Data1,{0.0, 1.0, 0.0, 0.8})
end


-- Used to refresh the Winning UI.
function addon:refresh()
    self:updateRecordsTab()
    self:updateStatsTab(addon.stats_selected_map_name)
end


function addon:updateRecordsTab()
    local wins, losses = self.dal:win_loss()
    local win_loss_ratio = self.dal:win_loss_ratio(wins, losses)

    Winning_UI_Info_Ratio_Wins:SetText(wins)
    Winning_UI_Info_Ratio_Losses:SetText(losses)
    Winning_UI_Info_Ratio_Value:SetText(win_loss_ratio)

    Winning_UI_Info_Stats_HKs_Value:SetText(
        self.dal:max_stat_in_ctx("HKs")
    )
    Winning_UI_Info_Stats_KBs_Value:SetText(
        self.dal:max_stat_in_ctx("KBs")
    )
    Winning_UI_Info_Stats_Damage_Value:SetText(self:render_damage(
        self.dal:max_stat_in_ctx("DamageRaw")
    ))
    Winning_UI_Info_Stats_Healing_Value:SetText(self:render_healing(
        self.dal:max_stat_in_ctx("HealingRaw")
    ))
    Winning_UI_Info_Stats_Streak_Value:SetText(
        self.dal:longest_streak()
    )

    -- Now update the actual table.
    self.records_table:SortData()
end

function addon:updateStatsTab(map_name)
    self:updateStatsMapsList(map_name)
    self:updateSelectedMapStats(map_name)
end

function addon:updateStatsMapsList(map_name)
    log("Updating stats map list.")
    for i, button in pairs(addon.buttons) do
        -- log("Checking map id: " .. tostring(i))
        local localizedName, canEnter, isHoliday, isRandom, battleGroundID, mapDescription, BGMapID, maxPlayers, gameType, iconTexture = GetBattlegroundInfo(i+1)

        button:Show()
        button.NameText:SetText(localizedName)

        local win_loss = self.dal:map_wins_losses(localizedName)
        button.WinLossDetail:SetText(format("%s - %s", self:green(win_loss.wins), self:red(win_loss.losses)))
        button.WinLossDetail:SetAlpha(0.5)

        button.RatioText:SetText(self.dal:win_loss_ratio(win_loss.wins, win_loss.losses))
        button.Icon:SetTexture(iconTexture or DEFAULT_BG_TEXTURE)
        button.map_name = localizedName

        if ( map_name == localizedName) then
            button.SelectedTexture:Show();
            button.NameText:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
            button.RatioText:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
        else
            button.SelectedTexture:Hide()
            button.NameText:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
            button.RatioText:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);
        end

        button:Show()
    end
end


function addon:updateSelectedMapStats(map_name)
    Winning_UI_Stats_Maps_Details_Info_Map_Label:SetText(map_name)
    if map_name then
        log("Updating selected map stats. " .. map_name)
        local bg_data = self.dal:daily_graph_data(map_name)

        log("Resetting stats graph data.")
        self.stats_graph:ResetData()

        log("Setting stats graph data for bg.")
        assert(bg_data.losses)
        assert(bg_data.wins)
        self.stats_graph:AddFilledDataSeries(bg_data.losses, {1.0, 0.0, 0.0, 1.0})
        self.stats_graph:AddFilledDataSeries(bg_data.wins, {0.0, 1.0, 0.0, 1.0})

        -- TODO: show graph frame.
        -- self.stats_graph:Show()?
    else
        log("No map selected for graph data.")
        -- TODO: hide graph frame.
        -- self.stats_graph:Hide()?
    end
end


-- Called after the DAL is done updating the data according to the context.
function addon:records_updated()
    self:refresh()
end


-- called when the context dropdowns are updated.
function addon:context_updated()
    self.dal:context_updated(self:current_context())
end


function addon:current_context()
    local ctx = {}

    -- Type context
    ctx.type = addon.db.faction.type_context_value
    log("Type context of: " .. tostring(ctx.type))
    
    -- Time frame context
    local time_ctx = addon.db.faction.time_context_value

    local time_value = self.dal.ALL
    if time_ctx == self.dal.SEASON then
        time_value = GetCurrentArenaSeason()
    elseif time_ctx == self.dal.SIX_HOURS then
        time_value = time() - (60*60*6)
    end

    log("Time ctx of: " .. tostring(time_ctx))
    log("Time value of: " .. tostring(time_value))

    ctx.time = {}
    ctx.time.context = time_ctx
    ctx.time.value = time_value


    -- Map context
    local map_ctx = addon.db.faction.map_context_value

    log("Map ctx of: " .. tostring(map_ctx))
    ctx.map = {}
    ctx.map.value = map_ctx


    -- Spec context
    ctx.spec = addon.db.faction.spec_context_value

    return ctx
end


-- Migrate data if necessary. Perform a migration for each version
function addon:migrate()
    if addon.db.global.data_version == 0 then
        log("Migrating data: 0 -> 1")
        addon.db.faction.bg_records = copy(addon.db.char.bg_records)
        addon.db.faction.all_records = copy(addon.db.char.all_records)
        addon.db.faction.bg_db_data = copy(addon.db.char.bg_db_data)
        addon.db.faction.type_context_value = copy(addon.db.char.type_context_value)
        addon.db.faction.time_context_value = copy(addon.db.char.time_context_value)
        addon.db.faction.map_context_value = copy(addon.db.char.map_context_value)
        addon.db.faction.graph_context_value = copy(addon.db.char.graph_context_value)
        addon.db.global.data_version = 1
    end
end


-- Setup specs for this faction data.
function addon:setupSpecs()
    log("Setting up specs.")
    self.all_specs = self.dal:get_specs()
    self.spec_items = {}
    local seen_specs = {}
    for i, v in pairs(self.all_specs) do
        local spec = {}
        if v == 0 then
            spec.label = L["All"]
            spec.spec = v
        else
            local id, name, description, icon, background, role = GetSpecializationInfoByID(v)
            spec.label = name
            spec.spec = id
        end
        table.insert(self.spec_items, spec)
    end

end


-- This stuff is similar to setup_ui, except that the Battleground info isn't
-- available at the time setup_ui is initially called. Just set it up here
-- after the addon has been loaded.
function addon:setupDisplay()

    -- Records Tab
    for i=1, GetNumBattlegroundTypes() do
        -- 1 == not implement CTF3 map
        -- 2 == bullshit temporary Southshore map.
        if i ~= 1  then
            local localizedName, canEnter, isHoliday, isRandom, BattleGroundID, mapDescription = GetBattlegroundInfo(i)
            addon.map_items[i] = localizedName
        else
            addon.map_items[1] = "All"
        end
    end

    -- Stats Tab
    if addon.buttons == nil then
        addon.buttons = {}
        local button
        local button_height = 45
        local current_offset = 0
        for i, v in pairs(addon.map_items) do
            if i ~= 1 then
                -- log("creating map button for: " .. v)
                local map_button_name = format("Winning_Map_Button_%s", i)
                button = CreateFrame("BUTTON", map_button_name, Winning_UI_Stats_Maps_List, "Winning_BG_Button_Tpl")
                if button then
                    -- log("Created button.")
                end
                -- log("setting button TOP to offset: " .. tostring(current_offset))
                button:SetPoint("TOP", Winning_UI_Stats_Maps_List, "TOP", 0, current_offset)
                button:SetWidth(309)
                button:SetHeight(button_height)
                current_offset = current_offset - button_height
                table.insert(addon.buttons, button)
            end
        end
    end
end
