local C = Winning_Constants


Winning_Chat_Icon_Texture = "\124TInterface\\PvPRankBadges\\PvPRank12:15:15:0:0\124t"
-- Insert our icon into chat messages.
Winning_Chat_Icon = function(msg)
    return format("%s %s", Winning_Chat_Icon_Texture, msg)
end


-- Convenience function for sending messages to the chat window.
Winning_Chat = function(msg)
    text = format("Winning: %s", msg)
    result = Winning_Chat_Icon(text)
	DEFAULT_CHAT_FRAME:AddMessage(result)
    return result
end


-- Logging function that takes addon debug state into account.
Winning_Log = function(msg)
	if C.ADDON_DEBUG then
        msg = format("[%s] - %s", date("%H:%M:%S", time()), msg)
        Winning_Chat(msg)
        return msg
	end
end


-- Convert seconds into a duration string: "34:01"
Winning_Duration = function(total_seconds)
    local seconds = math.floor(total_seconds % 60)
    local minutes = math.floor(total_seconds / 60)
    return format("%d:%s", minutes, tostring(seconds):rpad(2, "0"))
end


local random = math.random
Winning_UUID = function()
    local template ='xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'
    return string.gsub(template, '[xy]', function (c)
        local v = (c == 'x') and random(0, 0xf) or random(8, 0xb)
        return string.format('%x', v)
    end)
end


Winning_Humanify = function(num)
    num = tonumber(num) or 0
    local adjusted = 0
    if num > 999999999 then
        -- round billions
        return format("%.1fB", num/1000000000)
    elseif num > 999999 then
        -- round millions
        return format("%.2fM", num/1000000)
    elseif num > 999 then
        -- round thousands
        return format("%dk", math.ceil(num/1000))
    end
    return num
end


-- Stolen from the awesome BGSpy addon:
-- http://www.wowace.com/addons/bg-spy/
Winning_Parse_Name = function(identifier)
    realm = GetRealmName()
	-- Pattern assumes identifier is formatted as "name" or "name-server"
	-- Name is composed of every character except for '-'
	-- Server is any characters after an optional '-' character
	-- When no '-' character is present, the player is from the local realm
	local name, server = string.match(identifier, "^([^%-]+)-?(.*)")
	return name, server == "" and realm or server
end


-- Thanks, lua. Real nice.
function tcount(tbl)
    local count = 0
    for i, v in pairs(tbl) do
        count = count + 1
    end
    return count
end


-- Stolen from: https://gist.github.com/xytis/5361405
-- Print contents of `tbl`, with indentation.
-- `indent` sets the initial level of indentation.
function tprint (tbl, indent)
  if not indent then indent = 4 end
  for k, v in pairs(tbl) do
    formatting = string.rep("  ", indent) .. k .. ": "
    if type(v) == "table" then
      print(formatting)
      tprint(v, indent+1)
    else
      print(formatting .. tostring(v))
    end
  end
end

-- Pads str to length len with char from right
-- Mostly barrowed from here: http://snipplr.com/view/13092/strlpad--pad-string-to-the-left/
string.lpad = function(str, len, char)
    local s = str or "";
    if char == nil then char = ' ' end
    return s .. string.rep(char, len - #s)
end

string.rpad = function(str, len, char)
    local s = str or "";
    if char == nil then char = ' ' end
    return string.rep(char, len - #s) .. s
end


-- Copy values from char db to faction db, stolen from here:
-- http://stackoverflow.com/questions/640642/how-do-you-copy-a-lua-table-by-value
function copy(obj, seen)
  if type(obj) ~= 'table' then return obj end
  if seen and seen[obj] then return seen[obj] end
  local s = seen or {}
  local res = setmetatable({}, getmetatable(obj))
  s[obj] = res
  for k, v in pairs(obj) do res[copy(k, s)] = copy(v, s) end
  return res
end
